use crate::fqpath::*;
use crate::utils::{ident_prefix, ident_suffix};
use proc_macro2::{Ident, TokenStream};
use quote::{quote, quote_spanned, ToTokens};
use syn::{spanned::Spanned, Attribute, Data, DataStruct, DeriveInput, Field, Meta};

macro_rules! err_message {
    ($spannable: expr, $($tok:tt)*) => {
		return quote_spanned!(
			$spannable.span() =>
			compile_error!($($tok)*);
		)
	};
}

pub fn asset_system(
	DeriveInput {
		vis,
		data,
		ident,
		attrs,
		..
	}: DeriveInput,
) -> TokenStream {
	let struct_data = match data {
		Data::Struct(data_struct) => data_struct,
		Data::Enum(e) => err_message!(
			e.enum_token,
			"Can't create an asset system from an enum type"
		),
		Data::Union(u) => err_message!(
			u.union_token,
			"Can't create an asset system from a union type"
		),
	};

	let fields = define_fields(&struct_data);
	let getters = define_getters(&struct_data);
	let loader_properties = define_loader_properties(&attrs);
	let loader = define_loader(&ident, &struct_data, &loader_properties);

	quote! {
		#[derive(#FQDefault, #BevyResource)]
		#vis struct #ident {
			#fields
		}

		impl #ident {
			#getters
		}

		#loader
	}
}

fn define_fields(data: &DataStruct) -> TokenStream {
	data.fields
		.iter()
		.map(|Field { ident, ty, .. }| {
			quote! { pub #ident: #FQHashMap<#FQString, #BevyHandle<#ty>>, }
		})
		.collect()
}

fn define_getters(data: &DataStruct) -> TokenStream {
	data.fields
		.iter()
		.map(|Field { ident, ty, .. }| {
			let string_name = ident.to_token_stream().to_string();

			quote! {
				pub fn #ident<T: ToString>(&self, name: T) -> #BevyHandle<#ty> {
					let key = name.to_string();
					match self.#ident.get(&key) {
						Some(handle) => handle.clone_weak(),
						None => {
							let keys = self.#ident.keys();
							panic!(
								"\n\nTried to fetch {} asset with a missing key: {}.\nPossible keys: {}\n\n",
								#string_name,
								name.to_string(),
								keys.map(|k| format!("'{}'", k))
									.collect::<#FQVec<#FQString>>()
									.join(", ")
							)
						}
					}
				}
			}
		})
		.collect()
}

fn define_loader(ident: &Ident, data: &DataStruct, extra_props: &TokenStream) -> TokenStream {
	let loader_name = ident_suffix(ident, "Loader");
	let load_functions = define_load_functions(data);
	let load_generic = define_load_generic();

	quote! {
		#[derive(#BevySystemParam)]
		pub struct #loader_name<'w> {
			storage: #BevyResMut<'w, #ident>,
			server: #BevyRes<'w, #BevyAssetServer>,
			#extra_props
		}

		impl <'w>#loader_name<'w> {
			#load_generic
			#load_functions
		}
	}
}

fn define_load_generic() -> TokenStream {
	quote! {
		pub fn load_list<T, Load>(&mut self, list: #FQVec<(#FQString, #FQString)>, load: Load) -> #FQVec<#BevyHandle<T>>
		where
			T: #FQSend + #FQSync + #BevyAsset + 'static,
			Load: Fn(&mut Self, #FQString, #FQString) -> #BevyHandle<T>
		{
			list
				.iter()
				.map(|(path, key)| load(self, path.clone(), key.clone()))
				.collect()
		}
	}
}

fn define_load_functions(data: &DataStruct) -> TokenStream {
	data.fields
		.iter()
		.filter(|field| !field.attrs.iter().any(|attr| attr.path().is_ident("skip")))
		.map(|Field { ident, ty, .. }| {
			let some_ident = match ident {
				Some(id) => id,
				None => err_message!(ty, "Found asset system type without a property name"),
			};

			let load_single = ident_prefix(some_ident, "load_");
			let load_single_static = ident_suffix(&load_single, "_static");
			let load_list = ident_suffix(&load_single, "_list");
			let load_list_static = ident_suffix(&load_list, "_static");

			quote! {
				pub fn #load_single(&mut self, path: impl #FQDisplay, name: impl #FQDisplay) -> #BevyHandle<#ty> {
					let assets = vec![(path.to_string(), name.to_string())];
					let list = self.load_list(assets, |loader, path, key| {
						let handle: #BevyHandle<#ty> = loader.server.load(&path);
						loader.storage.#ident.insert(key, handle.clone());
						handle
					});

					list.first().expect("Failed to add asset").clone_weak()
				}
				pub fn #load_single_static(&mut self, path: &'static str, name: &'static str) -> #BevyHandle<#ty> {
					let assets = vec![(path.to_string(), name.to_string())];
					let list = self.load_list(assets, |loader, path, key| {
						let handle: #BevyHandle<#ty> = loader.server.load(&path);
						loader.storage.#ident.insert(key, handle.clone());
						handle
					});

					list.first().expect("Failed to add static asset").clone_weak()
				}
				pub fn #load_list(&mut self, assets: #FQVec<(#FQString, #FQString)>) -> #FQVec<#BevyHandle<#ty>> {
					self.load_list(assets, |loader, path, key| {
						let handle: #BevyHandle<#ty> = loader.server.load(&path);
						loader.storage.#ident.insert(key, handle.clone());
						handle
					})
				}
				pub fn #load_list_static(&mut self, assets: &'static [(&'static str, &'static str)]) -> #FQVec<#BevyHandle<#ty>> {
					let assets = assets.iter().map(|(a, b)| (a.to_string(), b.to_string())).collect();
					self.load_list(assets, |loader, path, key| {
						let handle: #BevyHandle<#ty> = loader.server.load(&path);
						loader.storage.#ident.insert(key, handle.clone());
						handle
					})
				}
			}
		})
		.collect()
}

fn define_loader_properties(attrs: &[Attribute]) -> TokenStream {
	attrs
		.iter()
		.filter(|attr| attr.path().is_ident("loader_property") && attr.meta.require_list().is_ok())
		.filter_map(|attr| {
			let list = match &attr.meta {
				Meta::List(list) => list,
				_ => return None,
			};
			let tokens = &list.tokens;
			Some(quote!(#tokens,))
		})
		.collect()
}
