use crate::fqpath::*;
use proc_macro2::{Span, TokenStream};
use quote::{quote, quote_spanned};
use syn::spanned::Spanned;
use syn::{Data, DeriveInput, Fields, FieldsNamed, FieldsUnnamed};

fn exactly_one<'a, T: Spanned>(mut it: impl Iterator<Item = &'a T>) -> Result<&'a T, Option<Span>> {
	let first = it.next();
	let second = it.next();

	if let Some(second) = second {
		Err(Some(second.span()))
	} else if let Some(first) = first {
		Ok(first)
	} else {
		Err(None)
	}
}

fn tuple_variant(input: &DeriveInput, fields: &FieldsUnnamed) -> TokenStream {
	match exactly_one(fields.unnamed.iter()) {
		Ok(field) => {
			let base_type = &input.ident;
			let field_type = &field.ty;

			quote! {
				#[automatically_derived]
				impl #FQFrom<#field_type> for #base_type {
					fn from(value: #field_type) -> #base_type {
						#base_type(value)
					}
				}
			}
		}
		Err(Some(field)) => {
			quote_spanned!(field.span() => compile_error!("Can't derive 'From' for a tuple struct with more than one inner property");)
		}
		Err(None) => {
			quote_spanned!(input.ident.span() => compile_error!("Must have at least one property to derive 'From'");)
		}
	}
}

fn named_variant(input: &DeriveInput, fields: &FieldsNamed) -> TokenStream {
	match exactly_one(fields.named.iter()) {
		Ok(field) => {
			let base_type = &input.ident;
			let field_type = &field.ty;
			let name = &field.ident;

			quote! {
				#[automatically_derived]
				impl #FQFrom<#field_type> for #base_type {
					fn from(value: #field_type) -> #base_type {
						#base_type {
							#name: value,
						}
					}
				}
			}
		}
		Err(Some(field)) => {
			quote_spanned!(field.span() => compile_error!("Can't derive 'From' for a tuple struct with more than one inner property");)
		}
		Err(None) => {
			quote_spanned!(input.ident.span() => compile_error!("Must have at least one property to derive 'From'");)
		}
	}
}

pub fn derive(input: DeriveInput) -> TokenStream {
	match &input.data {
		Data::Struct(at) => match &at.fields {
			Fields::Unnamed(fields) => tuple_variant(&input, fields),
			Fields::Named(fields) => named_variant(&input, fields),
			Fields::Unit => {
				quote_spanned!(input.ident.span() => compile_error!("'From' struct must contain at least one field");)
			}
		},
		Data::Enum(en) => {
			quote_spanned!(en.enum_token.span() => compile_error!("Deriving 'From' for enum is not supported");)
		}
		Data::Union(un) => {
			quote_spanned!(un.union_token.span() => compile_error!("Deriving 'From' for union is not supported");)
		}
	}
}
