use proc_macro2::{Ident, TokenStream, TokenTree};
use quote::{quote, quote_spanned, ToTokens};
use std::collections::HashMap;
use std::fmt::Display;
use syn::{
	spanned::Spanned, Attribute, Data, DataStruct, DeriveInput, Field, Fields, FieldsNamed, Meta,
	MetaNameValue, Visibility,
};

/// Convert a list of characters to snake case
pub fn snake_case(chars: impl Iterator<Item = char>) -> String {
	let mut acc = String::new();
	let mut prev = '_';
	for ch in chars {
		if ch.is_uppercase() && prev != '_' {
			acc.push('_');
		}
		acc.push(ch);
		prev = ch;
	}
	acc.to_lowercase()
}

/// Convert an identifier name to snake case
pub fn snake_case_ident(ident: &Ident) -> Ident {
	Ident::new(&snake_case(ident.to_string().chars()), ident.span())
}

pub fn ident_suffix(ident: &Ident, suffix: impl Display) -> Ident {
	Ident::new(&format!("{}{}", ident, suffix), ident.span())
}
pub fn ident_prefix(ident: &Ident, prefix: impl Display) -> Ident {
	Ident::new(&format!("{}{}", prefix, ident), ident.span())
}

macro_rules! to_str {
	($expr: expr) => {
		$expr.to_token_stream().to_string()
	};
}

#[allow(unused)]
pub fn unquote(value: String) -> String {
	match (value.starts_with('"'), value.ends_with('"')) {
		(true, true) => value[1..value.len() - 1].to_string(),
		(false, true) => value[..value.len() - 1].to_string(),
		(true, false) => value[1..].to_string(),
		(false, false) => value,
	}
}

pub fn find_asset_id(data: &DataStruct) -> Result<Field, Option<TokenStream>> {
	let mut found = None;
	for field in &data.fields {
		if field.ident.as_ref().map(|id| format!("{}", id)) == Some("id".into()) {
			if found.is_some() {
				return Err(Some(quote_spanned!(
					field.span() =>
					compile_error!("Asset must have only one of: property named 'id', property labelled with 'asset_id' attribute");
				)));
			} else {
				found = Some(field);
			}
			continue;
		}

		for attr in &field.attrs {
			if attr.path().is_ident("asset_id") {
				if found.is_some() {
					return Err(Some(quote_spanned!(
						field.span() =>
						compile_error!("Asset must have only one of: property named 'id', property labelled with 'asset_id' attribute");
					)));
				} else {
					found = Some(field);
				}
			}
		}
	}

	found.cloned().ok_or_else(|| None)
}

pub fn find_attribute(attrs: &Vec<Attribute>, path: impl ToString) -> Option<Attribute> {
	let attr_name = path.to_string();
	for attr in attrs {
		if attr.path().is_ident(&attr_name) {
			return Some(attr.clone());
		}
	}
	None
}

pub fn simple_attr_kv(attr: &Attribute) -> HashMap<String, String> {
	let mut attr_kv = HashMap::new();

	match &attr.meta {
		Meta::NameValue(mnv) => {
			attr_kv.insert(to_str!(mnv.path), to_str!(mnv.value));
		}
		Meta::List(ml) => {
			let mut buffer = Vec::with_capacity(4);
			for tok in ml.tokens.clone() {
				match tok {
					TokenTree::Punct(punct) if punct.as_char() == ',' => {
						let tokens = std::mem::take(&mut buffer)
							.iter()
							.fold(String::default(), |access, current| {
								format!("{}{}", access, current)
							});

						let mnv = tokens
							.parse::<TokenStream>()
							.expect("Invalid list of tokens for attribute");

						let mnv = syn::parse::<MetaNameValue>(mnv.into())
							.expect("Invalid list of tokens for attribute");

						attr_kv.insert(to_str!(mnv.path), to_str!(mnv.value));
					}
					other => {
						buffer.push(other);
					}
				}
			}

			if !buffer.is_empty() {
				let tokens = std::mem::take(&mut buffer)
					.iter()
					.fold(String::default(), |access, current| {
						format!("{}{}", access, current)
					});

				let mnv = tokens
					.parse::<TokenStream>()
					.expect("Invalid list of tokens for attribute");

				let mnv = syn::parse::<MetaNameValue>(mnv.into())
					.expect("Invalid list of tokens for attribute");

				attr_kv.insert(to_str!(mnv.path), to_str!(mnv.value));
			}
		}
		Meta::Path(pt) => {
			attr_kv.insert(to_str!(pt), String::new());
		}
	}

	attr_kv
}

#[allow(unused)]
pub fn module_wrapped(
	vis: &Visibility,
	mod_name: impl ToTokens,
	inner: TokenStream,
) -> TokenStream {
	quote! {
		mod #mod_name {
			#inner
		}
		#vis use #mod_name::*;
	}
}

pub fn only_struct<'a>(input: &'a DeriveInput) -> Result<&'a DataStruct, TokenStream> {
	match &input.data {
		Data::Struct(at) => Ok(at),
		Data::Enum(en) => Err(
			quote_spanned!(en.enum_token.span() => compile_error!("Unsupported for Enum, only for Struct");),
		),
		Data::Union(un) => Err(
			quote_spanned!(un.union_token.span() => compile_error!("Unsupported for Union, only for Struct");),
		),
	}
}

pub fn only_named_struct(data: &DataStruct) -> Result<&FieldsNamed, TokenStream> {
	match &data.fields {
		Fields::Named(named) => Ok(named),
		Fields::Unnamed(uf) => Err(
			quote_spanned!(uf.paren_token.span => compile_error!("Unsupported for tuple structs, must use named structs");),
		),
		Fields::Unit => Err(
			quote_spanned!(data.semi_token.span() => compile_error!("Unsupported for unit structs, must use named structs");),
		),
	}
}

#[cfg(test)]
mod tests {
	use crate::utils::{snake_case, unquote};
	use test_case::test_case;

	#[test_case("MyIdentifier", "my_identifier")]
	#[test_case("already_snaked", "already_snaked")]
	#[test_case("Some_Identifier", "some_identifier")]
	fn it_converts_identifiers(input: &str, expected: &str) {
		assert_eq!(snake_case(input.chars()), expected);
	}

	#[test_case(r#""Foo""#, "Foo")]
	#[test_case(r#""Left only"#, "Left only")]
	#[test_case(r#"Right only""#, "Right only")]
	#[test_case("No Quotes", "No Quotes")]
	#[test_case(r#"Middle " Unaffected"#, r#"Middle " Unaffected"#)]
	fn it_removes_quotes(input: &str, expected: &str) {
		assert_eq!(unquote(input.to_string()), expected);
	}
}
